import functools
import json
import logging
from collections import namedtuple
from concurrent.futures import CancelledError
from typing import Dict


def safe_loop():
    """Run coroutine in a safe loop.

    The coroutine runs in a 'while True' loop
    and exceptions logged.
    """

    def wrapper(func):
        @functools.wraps(func)
        async def wrapped(*args):
            logger = args[0].logger
            while True:
                try:
                    res = await func(*args)
                    if res is True:
                        break
                except Exception:
                    logger.error("Unkown exception", exc_info=True)

        return wrapped

    return wrapper



import io
import os
import sys
from argparse import ArgumentParser, FileType, Namespace
from typing import List

from owl_server.log import initlog
from owl_server.scripts import run_api, run_pipeline, run_scheduler


def parse_args(input: List[str]) -> Namespace:
    """Parse command line arguments.

    Parameters
    ----------
    input
        list of command line arguments

    Returns
    -------
    parsed arguments
    """
    parser = ArgumentParser()
    subparsers = parser.add_subparsers()

    # Scheduler
    scheduler = subparsers.add_parser("scheduler")
    scheduler.add_argument("--conf", required=False, type=FileType("r"))
    scheduler.set_defaults(func=run_scheduler)

    # Run the API
    api = subparsers.add_parser("api")
    api.add_argument("--conf", required=False, type=FileType("r"))
    api.set_defaults(func=run_api)

    # Run a pipeline
    run = subparsers.add_parser("pipeline")
    run.add_argument(
        "conf", nargs="?", type=FileType("r"), default=io.StringIO(os.getenv("PIPEDEF"))
    )
    run.set_defaults(func=run_pipeline)

    args = parser.parse_args(input)
    if not hasattr(args, "func"):
        parser.print_help()

    return args


def main():
    """Main entry point for owl.

    Invoke the command line help with::

        $ owl-server --help

    """
    initlog("cli")

    args = parse_args(sys.argv[1:])

    if hasattr(args, "func"):
        args.func(args)



import json
import logging.config
import logging.handlers
import os
from contextlib import suppress

import coloredlogs
import yaml
import zmq

from . import utils  # noqa: F401

logconf = {
    "api": """
version: 1
disable_existing_loggers: False
filters:
  filter_api:
    (): owl_server.log.LogFilter
    topic: API
handlers:
  console:
    class: owl_server.log.StreamHandler
    formatter: standard
    stream: 'ext://sys.stderr'
    filters: ["filter_api"]
  file:
    class: owl_server.log.TimeRotatingFileHandler
    filename: /var/run/owl/logs/api.log
    formatter: standard
    when: "d"
    backupCount: 7
    filters: ["filter_api"]
formatters:
  standard:
    class: owl_server.log.ColoredFormatter
    format: "%(asctime)s %(topic)s %(levelname)s %(name)s %(funcName)s - %(message)s"
loggers:
  owl.cli:
    handlers: [console]
    level: ${LOGLEVEL}
    propagate: false
  root:
    handlers: [console]
    level: INFO
    propagate: false
  uvicorn.error:
    handlers: [console, file]
    level: ${LOGLEVEL}
    propagate: false
  uvicorn.access:
    handlers: [console, file]
    level: ${LOGLEVEL}
    propagate: false
  uvicorn:
    handlers: [console, file]
    level: ${LOGLEVEL}
    propagate: false
""",
    "scheduler": """
version: 1
filters:
  filter_scheduler:
    (): owl_server.log.LogFilter
    topic: SCHEDULER
handlers:
  console:
    class: owl_server.log.StreamHandler
    formatter: standard
    stream: 'ext://sys.stderr'
    filters: ["filter_scheduler"]
  file:
    class: owl_server.log.TimeRotatingFileHandler
    filename: /var/run/owl/logs/scheduler.log
    formatter: standard
    when: "d"
    backupCount: 7
    filters: ["filter_scheduler"]
formatters:
  standard:
    class: owl_server.log.ColoredFormatter
    format: "%(asctime)s %(topic)s %(levelname)s %(name)s %(funcName)s - %(message)s"
loggers:
  owl.daemon.scheduler:
    handlers: [console, file]
    level: ${LOGLEVEL}
    propagate: false
  owl.cli:
    handlers: [console, file]
    level: ${LOGLEVEL}
    propagate: false
  root:
    handlers: [console, file]
    level: INFO
    propagate: false
""",
    "cli": """
version: 1
disable_existing_loggers: False
filters:
  filter_console:
    (): owl_server.log.LogFilter
    topic: CONSOLE
handlers:
  console:
    class: owl_server.log.StreamHandler
    formatter: standard
    stream: 'ext://sys.stderr'
    filters: ["filter_console"]
formatters:
  standard:
    class: owl_server.log.ColoredFormatter
    format: "%(asctime)s %(topic)s %(levelname)s %(name)s %(funcName)s - %(message)s"
loggers:
  owl.cli:
    handlers: [console]
    level: ${LOGLEVEL}
    propagate: false
""",
    "pipeline": """
version: 1
disable_existing_loggers: False
filters:
  filter_pipeline:
    (): owl_server.log.LogFilter
    topic: PIPELINE
    jobid: ${JOBID}
handlers:
  console:
    class: owl_server.log.StreamHandler
    formatter: standard
    stream: 'ext://sys.stderr'
    filters: ["filter_pipeline"]
  file:
    class: owl_server.log.TimeRotatingFileHandler
    filename: /tmp/pipeline.log
    formatter: standard
    when: "d"
    backupCount: 7
    filters: ["filter_pipeline"]
  zmq:
    class: owl_server.log.PUBHandler
    address: tcp://owl-scheduler:7002
    formatter: standard
    filters: ["filter_pipeline"]
formatters:
  standard:
    class: owl_server.log.ColoredFormatter
    format: "%(asctime)s %(topic)s %(levelname)s %(name)s %(funcName)s - %(message)s"
loggers:
  owl.daemon.pipeline:
    handlers: [console, file, zmq]
    level: ${LOGLEVEL}
    propagate: false
  owl.cli:
    handlers: [console, file, zmq]
    level: ${LOGLEVEL}
    propagate: false
  distributed:
    handlers: [console, file, zmq]
    level: INFO
    propagate: false
  root:
    handlers: [console, file, zmq]
    level: INFO
    propagate: false
""",
}


def initlog(val):
    """Configure logging."""
    log_config = yaml.safe_load(logconf[val])
    logging.config.dictConfig(log_config)
    return log_config


def logging_iteractive():
    """Helper function for iteractive sessions."""
    log_config = yaml.safe_load(logconf)
    initlog(log_config)


class LogFilter(logging.Filter):
    def __init__(self, **kwargs):
        self.kwargs = kwargs

    def filter(self, record):
        flag = True
        for k, v in self.kwargs.items():
            try:
                flag = flag and (getattr(record, k) == v)
            except AttributeError:
                setattr(record, k, v)
        return flag


class StreamHandler(logging.StreamHandler):
    def __init__(self, stream=None):
        super().__init__(stream)


class TimeRotatingFileHandler(logging.handlers.TimedRotatingFileHandler):
    """Time rotating file handler.

    It makes the base directory if it does not exist.

    Parameters
    ----------
    filename
        name of file to log to
    """

    def __init__(self, filename: str, **kwargs):
        path = os.path.dirname(filename)
        os.makedirs(path, exist_ok=True)
        super().__init__(filename, **kwargs)

    def open(self):
        super()._open()


class PipelineFileHandler(logging.FileHandler):
    """Time rotating file handler.

    It makes the base directory if it does not exist.

    Parameters
    ----------
    filename
        name of file to log to
    """

    def __init__(self, filename: str, **kwargs):
        path = os.path.dirname(filename)
        os.makedirs(path, exist_ok=True)
        super().__init__(filename, **kwargs)

    def open(self):
        super()._open()


class ColoredFormatter(coloredlogs.ColoredFormatter):
    """Colored formatter for logs.

    Parameters
    ----------
    fmt
        log formatting string
    datefmt
        format string for dates
    style
        style of log format
    """

    def __init__(self, fmt: str, datefmt: str, style: str):
        super().__init__(fmt, datefmt, style=style)

    def format(self, record):
        try:
            record.topic
        except AttributeError:
            record.topic = ""
        return super().format(record)


class PUBHandler(logging.Handler):
    def __init__(self, address):
        super().__init__()
        self.address = address
        self.socket = None

    def createSocket(self):
        with suppress(zmq.ZMQError):
            self.ctx = zmq.Context()
            self.socket = self.ctx.socket(zmq.PUB)
            self.socket.connect(self.address)

    def makeJSON(self, record):
        ei = record.exc_info
        if ei:
            # just to get traceback text into record.exc_text ...
            self.format(record)
        d = dict(record.__dict__)
        d["msg"] = record.getMessage()
        d["args"] = None
        d["exc_info"] = None
        # delete 'message' if present: redundant with 'msg'
        d.pop("message", None)
        return json.dumps(d)

    def send(self, s, topic):
        if self.socket is None:
            self.createSocket()
        if self.socket:
            topic = f"{topic}".encode("utf-8")
            msg = s.encode("utf-8")
            self.socket.send_multipart([topic, msg])

    def handleError(self, record):
        self.socket.close()
        self.socket = None

    def emit(self, record):
        try:
            s = self.makeJSON(record)
            self.send(s, record.topic)
        except Exception:
            self.handleError(record)

    def close(self):
        if self.socket:
            self.socket.close()
            self.socket = None
            self.release()



"""

The database migration is based on alembic_. The usage is:

.. code-block:: text

    usage: owl-migrate [-h] [-c CONF]
                           {current,edit,show,history,revision,upgrade} ...

    Minerva database migration

    optional arguments:
      -h, --help            show this help message and exit

    command:
      {current,edit,show,history,revision,upgrade}
        current             Display current revision
        edit                Edit revision
        show                Show revision
        history             Show revision history
        revision            Create a new revision file
        upgrade             Upgrade to a later version


.. _alembic: http://alembic.zzzcomputing.com/en/latest/

"""

import argparse
from datetime import datetime

from alembic import command as alembic_commands
from alembic.config import Config

from owl_server.config import config

def main():
    main_parser = argparse.ArgumentParser(description='Owl database migration')
    cmd_subparsers = main_parser.add_subparsers(title="command",
                                                dest="command")

    # current
    current_parser = cmd_subparsers.add_parser("current",
                                               help="Display current revision")
    current_parser.add_argument('-v', '--verbose',
                                action='store_true',
                                help='More verbose.')

    # edit
    edit_parser = cmd_subparsers.add_parser("edit",
                                            help="Edit revision")
    edit_parser.add_argument('revision', type=str,
                             help='Revision')

    # show
    show_parser = cmd_subparsers.add_parser("show",
                                            help="Show revision")
    show_parser.add_argument('revision', type=str,
                             help='Revision')

    # history
    history_parser = cmd_subparsers.add_parser("history",
                                               help="Show revision history")
    history_parser.add_argument('-v', '--verbose',
                                action='store_true',
                                help='More verbose.')

    # revision
    revision_parser = cmd_subparsers.add_parser("revision",
                                                help="Create a new revision file")
    revision_parser.add_argument('-m', '--message',
                                 type=str,
                                 help='Revision message.')
    revision_parser.add_argument('--id',
                                 type=str,
                                 default=None,
                                 help='Revision id.')
    revision_parser.add_argument('--autogenerate',
                                 action='store_true',
                                 help='Auto generate SQL from database')

    # upgrade
    upgrade_parser = cmd_subparsers.add_parser("upgrade",
                                               help="Upgrade to a later version")
    upgrade_parser.add_argument('revision', type=str,
                                help='Revision')
    upgrade_parser.add_argument('--sql',
                                action='store_true',
                                help='SQL mode.')

    args = main_parser.parse_args()

    alembic_cfg = Config()
    alembic_cfg.set_main_option("script_location", "owl_server:migrations")
    alembic_cfg.set_main_option("sqlalchemy.url",
                                config.dbi)


    if args.command == 'current':
        alembic_commands.current(alembic_cfg, verbose=args.verbose)
    elif 'edit' in args.command:
        alembic_commands.edit(alembic_cfg, args.revision)
    elif 'show' in args.command:
        alembic_commands.show(alembic_cfg, args.revision)
    elif 'history' in args.command:
        alembic_commands.history(alembic_cfg, verbose=args.verbose)
    elif 'revision' in args.command:
        if args.id is None:
            rev_id = datetime.now().strftime('%Y%m%d')
        else:
            rev_id = args.id
        alembic_commands.revision(alembic_cfg,
                                  message=args.message,
                                  rev_id=rev_id,
                                  autogenerate=args.autogenerate)
    elif 'upgrade' in args.command:
        alembic_commands.upgrade(alembic_cfg,
                                 args.revision,
                                 sql=args.sql)




import asyncio
import logging
import logging.config
import signal
from argparse import Namespace
from pathlib import Path

from owl_server.config import config
from owl_server.daemon import Scheduler
from owl_server.log import initlog

logger = logging.getLogger("owl.cli")


def run_scheduler(args: Namespace) -> None:
    """Run scheduler.

    Parameters
    ----------
    arg
        Argparse namespace containing command line flags.
    """
    initlog("scheduler")

    for d in ["conf", "logs"]:
        path = Path("/var/run/owl") / d
        path.mkdir(parents=True, exist_ok=True)

    try:
        start_loop()
    except:  # noqa
        logger.critical("Error starting the scheduler", exc_info=True)


def start_loop():
    loop = asyncio.get_event_loop()
    logger.info("Starting Owl Scheduler")
    scheduler = Scheduler()

    loop.add_signal_handler(signal.SIGTERM, loop.stop)
    loop.add_signal_handler(signal.SIGINT, loop.stop)

    loop.run_forever()
    loop.run_until_complete(scheduler.stop())

    loop.close()



import hashlib
import logging
import os
import random
import string
import sys
import time

import yaml
from kubernetes_asyncio import client, config, utils
from kubernetes_asyncio.client.api_client import ApiClient
from kubernetes_asyncio.client.models.v1_persistent_volume_claim import \
    V1PersistentVolumeClaim
from kubernetes_asyncio.client.rest import ApiException
from kubernetes_asyncio.config.config_exception import \
    ConfigException  # noqa: F401

from .utils import _make_spec_from_dict

log = logging.getLogger("owl.daemon.scheduler")


# Setup K8 configs
# config.load_kube_config()
# config.load_incluster_config()
# configuration = client.Configuration()
# api_instance = client.BatchV1Api()


def kube_delete_empty_pods(namespace=None, phase=None):
    """
    Pods are never empty, just completed the lifecycle.
    As such they can be deleted.
    Pods can be without any running container in 2 states:
    Succeeded and Failed. This call doesn't terminate Failed pods by default.
    """
    namespace = namespace or "default"
    phase = phase or "Succeeded"
    # The always needed object
    deleteoptions = client.V1DeleteOptions()
    # We need the api entry point for pods
    api_pods = client.CoreV1Api()
    # List the pods
    try:
        pods = api_pods.list_namespaced_pod(namespace, pretty=True, timeout_seconds=60)
    except ApiException as e:
        logging.error("Exception when calling CoreV1Api->list_namespaced_pod: %s\n" % e)

    for pod in pods.items:
        logging.debug(pod)
        podname = pod.metadata.name
        try:
            if pod.status.phase == phase:
                api_response = api_pods.delete_namespaced_pod(
                    podname, namespace, deleteoptions
                )
                logging.info("Pod: {} deleted!".format(podname))
                logging.debug(api_response)
            else:
                logging.info(
                    "Pod: {} still not done... Phase: {}".format(
                        podname, pod.status.phase
                    )
                )
        except ApiException as e:
            logging.error(
                "Exception when calling CoreV1Api->delete_namespaced_pod: %s\n" % e
            )

    return


def kube_cleanup_finished_jobs(namespace=None, state=None):
    """
    Since the TTL flag (ttl_seconds_after_finished) is still in alpha (Kubernetes 1.12) jobs need to be cleanup manually
    As such this method checks for existing Finished Jobs and deletes them.
    By default it only cleans Finished jobs. Failed jobs require manual intervention or a second call to this function.
    Docs: https://kubernetes.io/docs/concepts/workloads/controllers/jobs-run-to-completion/#clean-up-finished-jobs-automatically
    For deletion you need a new object type! V1DeleteOptions! But you can have it empty!
    CAUTION: Pods are not deleted at the moment. They are set to not running, but will count for your autoscaling limit, so if
             pods are not deleted, the cluster can hit the autoscaling limit even with free, idling pods.
             To delete pods, at this moment the best choice is to use the kubectl tool
             ex: kubectl delete jobs/JOBNAME.
             But! If you already deleted the job via this API call, you now need to delete the Pod using Kubectl:
             ex: kubectl delete pods/PODNAME
    """
    # deleteoptions = client.V1DeleteOptions()
    namespace = namespace or "default"
    state = state or "Finished"
    try:
        jobs = api_instance.list_namespaced_job(
            namespace, pretty=True, timeout_seconds=60
        )
        # print(jobs)
    except ApiException as e:
        print("Exception when calling BatchV1Api->list_namespaced_job: %s\n" % e)

    # Now we have all the jobs, lets clean up
    # We are also logging the jobs we didn't clean up because they either failed or are still running
    for job in jobs.items:
        logging.debug(job)
        jobname = job.metadata.name
        jobstatus = job.status.conditions
        if job.status.succeeded == 1:
            # Clean up Job
            logging.info(
                "Cleaning up Job: {}. Finished at: {}".format(
                    jobname, job.status.completion_time
                )
            )
            try:
                # What is at work here. Setting Grace Period to 0 means delete ASAP. Otherwise it defaults to
                # some value I can't find anywhere. Propagation policy makes the Garbage cleaning Async
                api_response = api_instance.delete_namespaced_job(
                    jobname,
                    namespace,
                    grace_period_seconds=0,
                    propagation_policy="Background",
                )
                logging.debug(api_response)
            except ApiException as e:
                print(
                    "Exception when calling BatchV1Api->delete_namespaced_job: %s\n" % e
                )
        else:
            if jobstatus is None and job.status.active == 1:
                jobstatus = "active"
            logging.info(
                "Job: {} not cleaned up. Current status: {}".format(jobname, jobstatus)
            )

    # Now that we have the jobs cleaned, let's clean the pods
    # kube_delete_empty_pods(namespace)
    # And we are done!
    return


def kube_create_job_object(
    name,
    container_image,
    command=None,
    namespace=None,
    container_name=None,
    env_vars=None,
    extraConfig=None,
    service_account_name=None,
):
    """
    Create a k8 Job Object
    Minimum definition of a job object:
    {'api_version': None, - Str
    'kind': None,     - Str
    'metadata': None, - Metada Object
    'spec': None,     -V1JobSpec
    'status': None}   - V1Job Status
    Docs: https://github.com/kubernetes-client/python/blob/master/kubernetes/docs/V1Job.md
    Docs2: https://kubernetes.io/docs/concepts/workloads/controllers/jobs-run-to-completion/#writing-a-job-spec
    Also docs are pretty pretty bad. Best way is to ´pip install kubernetes´ and go via the autogenerated code
    And figure out the chain of objects that you need to hold a final valid object So for a job object you need:
    V1Job -> V1ObjectMeta
          -> V1JobStatus
          -> V1JobSpec -> V1PodTemplate -> V1PodTemplateSpec -> V1Container

    Now the tricky part, is that V1Job.spec needs a .template, but not a PodTemplateSpec, as such
    you need to build a PodTemplate, add a template field (template.template) and make sure
    template.template.spec is now the PodSpec.
    Then, the V1Job.spec needs to be a JobSpec which has a template the template.template field of the PodTemplate.
    Failure to do so will trigger an API error.
    Also Containers must be a list!
    Docs3: https://github.com/kubernetes-client/python/issues/589
    """
    env_vars = env_vars or {}
    namespace = namespace or "default"
    command = command or "sleep 60"
    container_name = container_name or "jobcontainer"
    # Body is the object Body
    body = client.V1Job(api_version="batch/v1", kind="Job")
    # Body needs Metadata
    # Attention: Each JOB must have a different name!
    body.metadata = client.V1ObjectMeta(namespace=namespace, name=name)
    # And a Status
    body.status = client.V1JobStatus()
    # Now we start with the Template...
    template = client.V1PodTemplate()
    template.template = client.V1PodTemplateSpec()
    # Passing Arguments in Env:
    env_list = []
    for env_name, env_value in env_vars.items():
        env_list.append(client.V1EnvVar(name=env_name, value=str(env_value)))

    if (resources := extraConfig.resources) is not None:
        resourcesReq = client.V1ResourceRequirements(**resources)
    else:
        resourcesReq = None

    if (security_context := extraConfig.securityContext) is not None:
        secReq = _make_spec_from_dict(security_context, client.V1PodSecurityContext)
    else:
        secReq = None

    volume_mounts = []
    for vm in extraConfig.volumeMounts:
        volume_mounts.append(_make_spec_from_dict(vm, client.V1VolumeMount))

    container = client.V1Container(
        name=container_name,
        image=container_image,
        env=env_list,
        image_pull_policy="IfNotPresent",
        args=command.split(),
        resources=resourcesReq,
        volume_mounts=volume_mounts,
    )

    volumes = []
    for vol in extraConfig.volumes:
        volumes.append(_make_spec_from_dict(vol, client.V1Volume))

    template.template.spec = client.V1PodSpec(
        containers=[container],
        restart_policy="Never",
        termination_grace_period_seconds=30,
        volumes=volumes,
        service_account_name=service_account_name,
        security_context=secReq,
    )

    # And finaly we can create our V1JobSpec!
    body.spec = client.V1JobSpec(
        ttl_seconds_after_finished=300, template=template.template, backoff_limit=1
    )
    return body


async def kube_create_job(
    name,
    image,
    command=None,
    namespace=None,
    extraConfig=None,
    env_vars=None,
    service_account_name=None,
):
    namespace = namespace or "default"
    # Create the job definition
    body = kube_create_job_object(
        name,
        image,
        env_vars=env_vars,
        namespace=namespace,
        command=command or "sleep 60",
        extraConfig=extraConfig or {},
        service_account_name=service_account_name,
    )

    config.load_incluster_config()
    async with ApiClient() as api:
        v1 = client.BatchV1Api(api)
        res = await v1.create_namespaced_job(namespace, body, pretty=True)
    return res


async def kube_delete_job(name, namespace=None):
    config.load_incluster_config()
    async with ApiClient() as api:
        v1 = client.BatchV1Api(api)
        res = await v1.delete_namespaced_job(
            name, namespace or "default", propagation_policy="Background", grace_period_seconds=0
        )
    return res


def id_generator(size=12, chars=string.ascii_lowercase + string.digits):
    return "".join(random.choice(chars) for _ in range(size))


async def kube_test_credentials():
    """
    Testing function.
    If you get an error on this call don't proceed. Something is wrong on your connectivty to
    Google API.
    Check Credentials, permissions, keys, etc.
    Docs: https://cloud.google.com/docs/authentication/
    """
    config.load_incluster_config()

    async with ApiClient() as api:
        v1 = client.BatchV1Api(api)
        await v1.get_api_resources()



import asyncio
import concurrent.futures
import json
import logging
import os
from contextlib import suppress

import zmq
import zmq.asyncio
from dask.config import config as dask_config
from dask_kubernetes import KubeCluster
from owl_server import pipelines
from owl_server.config import config
from voluptuous import Invalid, MultipleInvalid

from .utils import safe_loop


class Pipeline:
    """Pipeline worker.

    The pipeline worker is started by the scheduler when a pipeline needs to
    be run. This worker runs the pipeline code, starting the swarm containers if
    necessart and is responsible of following the execution status and cleaning up.

    The process of running a pipeline is then as follows:

    1. The main scheduler starts a pipeline worker and sends the pipeline definition
       file and extra configuration needed.
    2. The worker loads the pipeline code and validates it against its schema.
    3. If running in swarm mode, the cluster starts the docker containers as requested.
    4. The worker starts a separate thread and runs the pipeline code.
    5. The main thread listens for heartbeat connections from the scheduler and waits for
       pipeline completion.
    6. The worker responds with the pipeline completion result to the scheduler.
    7. The scheduler stops the pipeline worker.

    Parameters
    ----------
    conf
        configuration
    logconf
        logging configuration
    """

    def __init__(self, pdef):
        self.logger = logging.getLogger("owl.daemon.pipeline")
        self.uid = os.environ.get("UID")
        self.executor = concurrent.futures.ThreadPoolExecutor(max_workers=3)
        self.pdef = pdef
        self.name = self.pdef["name"]
        self.info = {}
        self._tasks = []
        self.is_started = False
        self.status = "STARTING"
        self.loop = asyncio.get_event_loop()
        self.loop.run_until_complete(self.start())

    async def start(self):
        self.logger.info("Starting pipeline ID %s with %s", self.uid, self.pdef)
        self.is_started = True

        await self._setup_sockets()

        self._tasks.append(asyncio.ensure_future(self.heartbeat()))

        try:
            await self.loop.run_in_executor(self.executor, self.check_pipeline)
        except (AttributeError, Invalid, MultipleInvalid) as e:
            self.logger.critical(
                "Unable to load pipeline %s: %s", self.name, e, exc_info=True
            )
            self.status = "ERROR"
            return

        await self.start_dask_cluster()

        func = getattr(pipelines, self.name)
        self.logger.info(f"Running {func!r} with {self.pdef}")
        self.proc = self.loop.run_in_executor(
            self.executor, func, self.pdef, self.cluster
        )
        self.proc.add_done_callback(self.pipeline_done)
        self._tasks.append(self.proc)

        self.status = "RUNNING"

    def pipeline_done(self, future: asyncio.Future):
        self.logger.info("Pipeline ID %s finished", self.uid)
        if (e := future.exception()):
            self.logger.critical("Failed to run pipeline %s", e)
            self.status = "ERROR"
        else:
            self.status = "FINISHED"

    async def start_dask_cluster(self):
        self.logger.info("Starting Dask cluster")
        self.dask_config()
        nworkers = self.pdef["resources"]["workers"]
        self.cluster = await KubeCluster(asynchronous=True, n_workers=nworkers)
        # self.cluster.adapt(minimum=nworkers, maximum=nworkers)
        self.logger.debug("Scheduler address: %s", self.cluster.scheduler_address)

    def dask_config(self):
        resources = self.pdef["resources"]
        worker = config.dask.kubernetes['worker-template']['spec']['containers'][0]
        nthreads = resources['threads']
        nprocs = resources['processes']
        memory = resources['memory']
        args = [
            "dask-worker",
            "--nthreads",
            f"{nthreads}",
            "--nprocs",
            f"{nprocs}",
            "--memory-limit",
            f"{memory}GB",
            "--death-timeout",
            "60",
        ]
        worker["args"] = args
        worker["resources"] = {
            "limits": {
                "cpu": f"{nthreads*nprocs}",
                "memory": f"{memory}G",
            },
            "requests": {
                "cpu": f"{nthreads*nprocs}",
                "memory": f"{memory}G",
            },
        }

        for k in ["name", "scheduler-template", "worker-template"]:
            dask_config["kubernetes"][k] = config.dask.kubernetes[k]

    async def _setup_sockets(self):
        self.owl_host = os.environ.get("OWL_SCHEDULER_SERVICE_HOST")
        self.pipe_port = os.environ.get("OWL_SCHEDULER_SERVICE_PORT_PIPE")
        self.pipe_addr = f"tcp://{self.owl_host}:{self.pipe_port}"
        self.logger.debug("Connecting to %s", self.pipe_addr)
        self.ctx = zmq.asyncio.Context()

        self.pipe_socket = self.ctx.socket(zmq.DEALER)
        self.pipe_socket.setsockopt(zmq.IDENTITY, self.uid.encode("utf-8"))
        self.pipe_socket.connect(self.pipe_addr)

        await asyncio.sleep(0)

    @safe_loop()
    async def heartbeat(self):
        msg = await self.pipe_socket.recv()
        self.logger.info("Pipeline %s: heartbeat received", self.uid)
        msg = {"status": self.status}
        await self.pipe_socket.send(json.dumps(msg).encode("utf-8"))

        if self.status not in ["PENDING", "RUNNING"]:
            with suppress(Exception):
                await self.cluster.close()
            # return True

    async def stop(self):
        """Stop pipeline worker.

        All tasks started by `start` are now cancelled.
        """
        self.logger.info("Stopping pipeline ID %s", self.uid)
        if self.is_started:
            for task in self._tasks:
                task.cancel()
                with suppress(asyncio.CancelledError):
                    await task

            with suppress(Exception):
                await self.cluster.close()

    def check_pipeline(self):
        self.logger.debug("Loading pipeline %s", self.name)
        self.func = getattr(pipelines, self.name)
        self.logger.debug("Loaded pipeline %s", self.name)
        self.info["version"] = self.func.__version__
        self.logger.debug("Checking pipeline schema")
        self.func.schema(self.pdef)

        



import base64
import hashlib
import math
import secrets
import warnings

RANDOM_STRING_CHARS = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789"


def get_random_string(length, allowed_chars=RANDOM_STRING_CHARS):
    """
    Return a securely generated random string.
    The bit length of the returned value can be calculated with the formula:
        log_2(len(allowed_chars)^length)
    For example, with default `allowed_chars` (26+26+10), this gives:
      * length: 12, bit length =~ 71 bits
      * length: 22, bit length =~ 131 bits
    """
    return "".join(secrets.choice(allowed_chars) for i in range(length))


def force_bytes(s, encoding="utf-8", errors="strict"):
    """
    Similar to smart_bytes, except that lazy instances are resolved to
    strings, rather than kept as lazy objects.
    If strings_only is True, don't convert (some) non-string-like objects.
    """
    # Handle the common case first for performance reasons.
    if isinstance(s, bytes):
        if encoding == "utf-8":
            return s
        else:
            return s.decode("utf-8", errors).encode(encoding, errors)
    return str(s).encode(encoding, errors)


def constant_time_compare(val1, val2):
    """Return True if the two strings are equal, False otherwise."""
    return secrets.compare_digest(force_bytes(val1), force_bytes(val2))


def pbkdf2(password, salt, iterations, dklen=0, digest=None):
    """Return the hash of password using pbkdf2."""
    if digest is None:
        digest = hashlib.sha256
    dklen = dklen or None
    password = force_bytes(password)
    salt = force_bytes(salt)
    return hashlib.pbkdf2_hmac(digest().name, password, salt, iterations, dklen)


class BasePasswordHasher:
    """
    Abstract base class for password hashers
    When creating your own hasher, you need to override algorithm,
    verify(), encode() and safe_summary().
    PasswordHasher objects are immutable.
    """

    algorithm = None
    library = None
    salt_entropy = 128

    def salt(self):
        """
        Generate a cryptographically secure nonce salt in ASCII with an entropy
        of at least `salt_entropy` bits.
        """
        # Each character in the salt provides
        # log_2(len(alphabet)) bits of entropy.
        char_count = math.ceil(self.salt_entropy / math.log2(len(RANDOM_STRING_CHARS)))
        return get_random_string(char_count, allowed_chars=RANDOM_STRING_CHARS)

    def verify(self, password, encoded):
        """Check if the given password is correct."""
        raise NotImplementedError(
            "subclasses of BasePasswordHasher must provide a verify() method"
        )

    def _check_encode_args(self, password, salt):
        if password is None:
            raise TypeError("password must be provided.")
        if not salt or "$" in salt:
            raise ValueError("salt must be provided and cannot contain $.")

    def encode(self, password, salt):
        """
        Create an encoded database value.
        The result is normally formatted as "algorithm$salt$hash" and
        must be fewer than 128 characters.
        """
        raise NotImplementedError(
            "subclasses of BasePasswordHasher must provide an encode() method"
        )

    def decode(self, encoded):
        """
        Return a decoded database value.
        The result is a dictionary and should contain `algorithm`, `hash`, and
        `salt`. Extra keys can be algorithm specific like `iterations` or
        `work_factor`.
        """
        raise NotImplementedError(
            "subclasses of BasePasswordHasher must provide a decode() method."
        )


class PBKDF2PasswordHasher(BasePasswordHasher):
    """
    Secure password hashing using the PBKDF2 algorithm (recommended)
    Configured to use PBKDF2 + HMAC + SHA256.
    The result is a 64 byte binary string.  Iterations may be changed
    safely but you must rename the algorithm if you change SHA256.
    """

    algorithm = "pbkdf2_sha256"
    iterations = 320000
    digest = hashlib.sha256

    def encode(self, password, salt, iterations=None):
        self._check_encode_args(password, salt)
        iterations = iterations or self.iterations
        hash = pbkdf2(password, salt, iterations, digest=self.digest)
        hash = base64.b64encode(hash).decode("ascii").strip()
        return "%s$%d$%s$%s" % (self.algorithm, iterations, salt, hash)

    def decode(self, encoded):
        algorithm, iterations, salt, hash = encoded.split("$", 3)
        assert algorithm == self.algorithm
        return {
            "algorithm": algorithm,
            "hash": hash,
            "iterations": int(iterations),
            "salt": salt,
        }

    def verify(self, password, encoded):
        decoded = self.decode(encoded)
        encoded_2 = self.encode(password, decoded["salt"], decoded["iterations"])
        return constant_time_compare(encoded, encoded_2)




from distributed import WorkerPlugin


class LoggingPlugin(WorkerPlugin):
    def __init__(self, config):
        self.config = config

    def setup(self, worker):
        import logging.config

        logging.config.dictConfig(self.config)




import logging
import os
import shutil
import sys
import traceback
from contextlib import suppress
from functools import wraps
from pathlib import Path
from typing import Callable, Dict, Optional

import pkg_resources
import voluptuous as vo
import yaml
from distributed import Client
from owl_server.log import logconf
from owl_server.plugins import LoggingPlugin

logger = logging.getLogger("owl.daemon.scheduler")

PIPELINES = dict()


def _import_pipelines():
    for e in pkg_resources.iter_entry_points("owl.pipelines"):
        f = e.load()
        try:
            schema = f.schema
        except AttributeError:
            schema = None
        PIPELINES[e.name] = register_pipeline(validate=schema)(f)


class register_pipeline:
    """Register a pipeline."""

    def __init__(self, *, validate: vo.Schema = None):
        self.schema = validate
        if self.schema is not None:
            self.schema.extra = vo.REMOVE_EXTRA

    def __call__(self, func: Callable):
        name = func.__name__

        @wraps(func)
        def wrapper(config: Dict, cluster=None):

            if self.schema is not None:
                _config = self.schema(config)
            else:
                _config = config

            if cluster is not None:
                try:
                    client = Client(cluster.scheduler_address)
                except Exception:
                    traceback_str = traceback.format_exc()
                    raise Exception(
                        "Error occurred. Original traceback " "is\n%s\n" % traceback_str
                    )
                c = yaml.safe_load(logconf["pipeline"])
                plugin = LoggingPlugin(c)
                client.register_worker_plugin(plugin)
            else:
                client = None
            try:
                func.main.config = config
                return func.main(**_config)
            except Exception:
                traceback_str = traceback.format_exc()
                raise Exception(
                    "Error occurred. Original traceback " "is\n%s\n" % traceback_str
                )
            finally:
                if client is not None:
                    client.close()
                logfile = os.environ.get("LOGFILE")
                with suppress(Exception):
                    path = Path(logfile)
                    shutil.copy(path, Path(func.main.config["output_dir"]) / path.name)

        PIPELINES[name] = wrapper
        return wrapper


def __getattr__(name: str):
    """Return a named plugin"""
    try:
        return PIPELINES[name]
    except KeyError:
        _import_pipelines()
        if name in PIPELINES:
            return PIPELINES[name]
        else:
            raise AttributeError(
                f"module {__name__!r} has no attribute {name!r}"
            ) from None


def __dir__():
    """List available plug-ins"""
    _import_pipelines()
    return list(PIPELINES.keys())



import json
import logging
import os
import re
import select
import sys
from collections import namedtuple
from typing import IO, Callable, Dict, Union

import yaml
from kubernetes import client

log = logging.getLogger("owl.cli")


SERIALIZATION_API_CLIENT = client.ApiClient()
SERIALIZATION_API_CLIENT.pool.close()

_path_matcher = re.compile(r"(\S+)?(\$\{([^}^{]+)\})")


def _path_constructor(loader, node):
    """Extract the matched value, expand env variable, and replace the match."""
    value = node.value
    match = _path_matcher.match(value)
    env_var = match.groups()[2]
    return value.replace(match.groups()[1], os.environ.get(env_var, ""))


yaml.add_implicit_resolver("!path", _path_matcher, None, yaml.SafeLoader)
yaml.add_constructor("!path", _path_constructor, yaml.SafeLoader)


def read_config(
    config: Union[str, IO[str]], validate: Callable = None
) -> Dict[str, str]:
    """Read configuration file.

    Parameters
    ----------
    config
        input configuration
    validate
        validation schema

    Returns
    -------
    parsed configuration
    """
    try:
        conf = yaml.safe_load(config)
    except:  # noqa
        log.critical("Unable to read configuration file.")
        raise
    if validate:
        conf = validate(conf)
    return conf


def read_config_stdin(*args):
    """Read config from stdin."""
    handler, *_ = select.select([sys.stdin], [], [], 1)
    if handler:
        config = yaml.safe_load(handler[0].readline())
    else:
        config = {}

    return config


_FakeResponse = namedtuple("_FakeResponse", ["data"])


def _make_spec_from_dict(_dict: Dict, kind=client.V1Pod):
    return SERIALIZATION_API_CLIENT.deserialize(
        _FakeResponse(data=json.dumps(_dict)), kind
    )




from argparse import Namespace

import uvicorn
import yaml
from owl_server.config import config
from uvicorn.config import LOGGING_CONFIG


def run_api(args: Namespace) -> None:  # pragma: nocover
    """Start the API service.

    Parameters
    ----------
    arg
        Argparse namespace containing command line flags.
    """
    from owl_server.api import app
    from owl_server.log import logconf

    LOGGING_CONFIG.update(yaml.safe_load(logconf["api"]))
    api_port = int(config.env.OWL_API_SERVICE_PORT)

    uvicorn.run(app, host="0.0.0.0", port=api_port)



from enum import unique

import sqlalchemy

metadata = sqlalchemy.MetaData()

Pipeline = sqlalchemy.Table(
    "pipeline",
    metadata,
    sqlalchemy.Column("id", sqlalchemy.Integer, primary_key=True),
    sqlalchemy.Column("config", sqlalchemy.JSON, nullable=False),
    sqlalchemy.Column("heartbeat", sqlalchemy.JSON, nullable=True),
    sqlalchemy.Column("user", sqlalchemy.String(length=32), nullable=False),
    sqlalchemy.Column(
        "status", sqlalchemy.String(length=16), default="PENDING", nullable=False
    ),
    sqlalchemy.Column("created_at", sqlalchemy.DateTime, nullable=False),
    sqlalchemy.Column("started_at", sqlalchemy.DateTime, nullable=True),
    sqlalchemy.Column("finished_at", sqlalchemy.DateTime, nullable=True),
    sqlalchemy.Column("elapsed", sqlalchemy.Float, nullable=True),
)

PipelineDefinition = sqlalchemy.Table(
    "pipeline_definition",
    metadata,
    sqlalchemy.Column("id", sqlalchemy.Integer, primary_key=True),
    sqlalchemy.Column("pdef", sqlalchemy.JSON, nullable=False),
    sqlalchemy.Column(
        "name", sqlalchemy.String(length=32), nullable=False, unique=True
    ),
    sqlalchemy.Column("extra_pip_packages", sqlalchemy.Text, nullable=False),
    sqlalchemy.Column("active", sqlalchemy.Boolean, default=True),
    # sqlalchemy.Column("version", sqlalchemy.String(length=32), nullable=False),
)

ContainerImage = sqlalchemy.Table(
    "container_image",
    metadata,
    sqlalchemy.Column("id", sqlalchemy.Integer, primary_key=True),
    sqlalchemy.Column("name", sqlalchemy.String(length=80), nullable=False),
    sqlalchemy.Column("version", sqlalchemy.String(length=32), nullable=False),
    sqlalchemy.Column("active", sqlalchemy.Boolean, default=True),
)

Token = sqlalchemy.Table(
    "token",
    metadata,
    sqlalchemy.Column("id", sqlalchemy.Integer, primary_key=True),
    sqlalchemy.Column("username", sqlalchemy.String(length=32), unique=True),
    sqlalchemy.Column("token", sqlalchemy.String(length=128)),
)

User = sqlalchemy.Table(
    "user",
    metadata,
    sqlalchemy.Column("id", sqlalchemy.Integer, primary_key=True),
    sqlalchemy.Column("username", sqlalchemy.String(length=32), unique=True),
    sqlalchemy.Column("password", sqlalchemy.String(length=32)),
    sqlalchemy.Column("is_admin", sqlalchemy.Boolean, default=False),
    sqlalchemy.Column("active", sqlalchemy.Boolean, default=True),
)



"""Initial revision

Revision ID: 20210804
Revises: 
Create Date: 2021-08-04 15:20:27.231408

"""
from os import urandom

import sqlalchemy as sa
from alembic import op
from owl_server.crypto import PBKDF2PasswordHasher

# revision identifiers, used by Alembic.
revision = "20210804"
down_revision = None
branch_labels = None
depends_on = None


def generate_temp_password(length):
    if not isinstance(length, int) or length < 8:
        raise ValueError("password must have positive length")
    chars = "ABCDEFGHJKLMNPQRSTUVWXYZ23456789"
    return "".join(chars[c % len(chars)] for c in urandom(length))


def insert_temp_password(user):
    temp_password = generate_temp_password(12)
    hasher = PBKDF2PasswordHasher()
    salt = hasher.salt()
    password = hasher.encode(temp_password, salt)
    cmd = 'INSERT INTO user VALUES (1, "{}", "{}", True, True)'.format(user, password)
    op.execute(cmd)
    with open("/var/run/owl/adminPassword", "w") as fh:
        fh.write(temp_password)


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "container_image",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("name", sa.String(length=80), nullable=False),
        sa.Column("version", sa.String(length=32), nullable=False),
        sa.Column("active", sa.Boolean(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_table(
        "pipeline",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("config", sa.JSON(), nullable=False),
        sa.Column("heartbeat", sa.JSON(), nullable=True),
        sa.Column("user", sa.String(length=32), nullable=False),
        sa.Column("status", sa.String(length=16), nullable=False),
        sa.Column("created_at", sa.DateTime(), nullable=False),
        sa.Column("started_at", sa.DateTime(), nullable=True),
        sa.Column("finished_at", sa.DateTime(), nullable=True),
        sa.Column("elapsed", sa.Float(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_table(
        "pipeline_definition",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("pdef", sa.JSON(), nullable=False),
        sa.Column("name", sa.String(length=32), nullable=False),
        sa.Column("extra_pip_packages", sa.Text(), nullable=False),
        sa.Column("active", sa.Boolean(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("name"),
    )
    op.create_table(
        "token",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("username", sa.String(length=32), nullable=True),
        sa.Column("token", sa.String(length=128), nullable=True),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("username"),
    )
    op.create_table(
        "user",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("username", sa.String(length=32), nullable=True),
        sa.Column("password", sa.String(length=32), nullable=True),
        sa.Column("is_admin", sa.Boolean(), nullable=True),
        sa.Column("active", sa.Boolean(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("username"),
    )
    insert_temp_password("admin")
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("user")
    op.drop_table("token")
    op.drop_table("pipeline_definition")
    op.drop_table("pipeline")
    op.drop_table("container_image")
    # ### end Alembic commands ###



import ast
import base64
import builtins
import json
import os
import re
import sys
import threading
import warnings
from collections.abc import Mapping

import yaml

from . import utils
from .schema import AttrDict, schema_server

paths = [
    os.getenv("OWL_ROOT_CONFIG", "/etc/owl"),
    os.path.join(sys.prefix, "etc", "owl"),
    os.path.join(os.path.expanduser("~"), ".config", "owl"),
    os.path.join(os.path.expanduser("~"), ".owl"),
]


if "OWL_CONFIG" in os.environ:
    PATH = os.environ["OWL_CONFIG"]
    paths.append(PATH)


global_config = config = AttrDict()

config_lock = threading.Lock()

defaults = []


def canonical_name(k, config):
    """Return the canonical name for a key.
    Handles user choice of '-' or '_' conventions by standardizing on whichever
    version was set first. If a key already exists in either hyphen or
    underscore form, the existing version is the canonical name. If neither
    version exists the original key is used as is.
    """
    try:
        if k in config:
            return k
    except TypeError:
        # config is not a mapping, return the same name as provided
        return k

    altk = k.replace("_", "-") if "_" in k else k.replace("-", "_")

    if altk in config:
        return altk

    return k


def update(old, new, priority="new"):
    """Update a nested dictionary with values from another
    This is like dict.update except that it smoothly merges nested values
    This operates in-place and modifies old

    Parameters
    ----------
    priority: string {'old', 'new'}
        If new (default) then the new dictionary has preference.
        Otherwise the old dictionary does.

    Examples
    --------
    >>> a = {'x': 1, 'y': {'a': 2}}
    >>> b = {'x': 2, 'y': {'b': 3}}
    >>> update(a, b)  # doctest: +SKIP
    {'x': 2, 'y': {'a': 2, 'b': 3}}
    >>> a = {'x': 1, 'y': {'a': 2}}
    >>> b = {'x': 2, 'y': {'b': 3}}
    >>> update(a, b, priority='old')  # doctest: +SKIP
    {'x': 1, 'y': {'a': 2, 'b': 3}}

    See Also
    --------
    dask.config.merge
    """
    for k, v in new.items():
        k = canonical_name(k, old)

        if isinstance(v, Mapping):
            if k not in old or old[k] is None:
                old[k] = AttrDict()
            update(old[k], v, priority=priority)
        elif priority == "new" or k not in old:
            old[k] = v

    return old


def merge(*dicts):
    """Update a sequence of nested dictionaries
    This prefers the values in the latter dictionaries to those in the former
    Examples
    --------
    >>> a = {'x': 1, 'y': {'a': 2}}
    >>> b = {'y': {'b': 3}}
    >>> merge(a, b)  # doctest: +SKIP
    {'x': 1, 'y': {'a': 2, 'b': 3}}
    See Also
    --------
    dask.config.update
    """
    result = {}
    for d in dicts:
        update(result, d)
    return result


def collect_yaml(paths=paths):
    """Collect configuration from yaml files
    This searches through a list of paths, expands to find all yaml or json
    files, and then parses each file.
    """
    # Find all paths
    file_paths = []
    for path in paths:
        if os.path.exists(path):
            if os.path.isdir(path):
                try:
                    file_paths.extend(
                        sorted(
                            [
                                os.path.join(path, p)
                                for p in os.listdir(path)
                                if os.path.splitext(p)[1].lower()
                                in (".json", ".yaml", ".yml")
                            ]
                        )
                    )
                except OSError:
                    # Ignore permission errors
                    pass
            else:
                file_paths.append(path)

    configs = []

    # Parse yaml files
    for path in file_paths:
        try:
            with open(path) as f:
                data = yaml.safe_load(f.read()) or {}
                configs.append(data)
        except OSError:
            # Ignore permission errors
            pass

    return configs


def collect(paths=paths, env=None):
    """
    Collect configuration from paths and environment variables
    Parameters
    ----------
    paths : List[str]
        A list of paths to search for yaml config files
    env : dict
        The system environment variables
    Returns
    -------
    config: dict
    See Also
    --------
    dask.config.refresh: collect configuration and update into primary config
    """
    if env is None:
        env = os.environ

    configs = collect_yaml(paths=paths)
    # configs.append(collect_env(env=env))

    return merge(*configs)


def refresh(config=config, defaults=defaults, **kwargs):
    """
    Update configuration by re-reading yaml files and env variables
    This mutates the global dask.config.config, or the config parameter if
    passed in.
    This goes through the following stages:
    1.  Clearing out all old configuration
    2.  Updating from the stored defaults from downstream libraries
        (see update_defaults)
    3.  Updating from yaml files and environment variables
    Note that some functionality only checks configuration once at startup and
    may not change behavior, even if configuration changes.  It is recommended
    to restart your python process if convenient to ensure that new
    configuration changes take place.
    See Also
    --------
    dask.config.collect: for parameters
    dask.config.update_defaults
    """
    config.clear()

    for d in defaults:
        update(config, d, priority="old")

    update(config, collect(**kwargs))


def update_defaults(new, config=config, defaults=defaults):
    """Add a new set of defaults to the configuration
    It does two things:
    1.  Add the defaults to a global collection to be used by refresh later
    2.  Updates the global config with the new configuration
        prioritizing older values over newer ones
    """
    defaults.append(new)
    update(config, new, priority="old")
    schema_server(config)


def _initialize():
    _defaults = {}

    update_defaults(_defaults)


refresh()
_initialize()



import asyncio
import json
import logging
import time
from contextlib import suppress
from pathlib import Path
from typing import Any, Dict

import aiohttp
import zmq
import zmq.asyncio
from aiohttp.client_exceptions import ClientConnectorError
from async_timeout import timeout
from owl_server import k8s
from owl_server.config import config
from owl_server.log import LogFilter, PipelineFileHandler

from .utils import safe_loop

MAX_PIPELINES = 6


class Scheduler:
    """Owl pipeline scheduler.

    The scheduler is responsible of communication between the user
    and the pipeline. It queues pipeline requests and runs them as
    resoures become available.
    """

    def __init__(self):
        self.logger = logging.getLogger("owl.daemon.scheduler")

        self.env = config.env  # enviroment variables from config
        self._token = config.pop(
            "token"
        )  # secret token to use in communication between API and OWL

        self.pipelines = {}  # list of pipelines running
        self._max_pipe = MAX_PIPELINES
        self.kube_metrics = {}
        config.pop("dbi")  # database configuration, remove it as it is not used here

        self.is_started = False
        self._tasks = []  # list of coroutines
        self.loop = asyncio.get_event_loop()
        self.loop.run_until_complete(self.start())

    @property
    def namespace(self) -> str:
        """Return namespace of running Pod

        Returns
        -------
        Name of namespace where the current pod is running.
        """
        if not hasattr(self, "_namespace"):
            self._namespace = open(
                "/var/run/secrets/kubernetes.io/serviceaccount/namespace"
            ).read()
        return self._namespace

    async def start(self):
        """Start scheduler.

        Start the following:

        - Check that we have correct Kubernetes credentials
        - Setup connection sockets.
        - HTTP client session for connections to the API.
        - Task for checking pipelines in the queue.
        - Task for checking pipelines status.
        - Task for receiving log messages.
        """
        if self.is_started:
            await asyncio.sleep(0)
            return

        self.logger.debug("Starting scheduler with %s", config)
        try:
            self.logger.debug("Testing Kubernetes credentials")
            await k8s.kube_test_credentials()
        except k8s.ConfigException as e:
            self.logger.critical("Unable to obtain Kubernetes configuration: %s", e)
            raise
        except k8s.ApiException as e:
            self.logger.critical("Exception when calling API: %s", e)
            raise

        self.session = aiohttp.ClientSession()

        try:
            await self._setup_sockets()
            await self._start_liveness_probe()
        except Exception as e:
            self.logger.critical("Unable to setup communication sockets: %s", e)

        self._start_tasks()

        self.is_started = True
        self.logger.info("Owl scheduler started")

    def _start_tasks(self):
        """Add tasks to the loop."""
        self._tasks.extend(
            [
                asyncio.create_task(self.load_pipelines()),
                asyncio.create_task(self.scheduler_heartbeat()),
                asyncio.create_task(self.status_pipelines()),
                asyncio.create_task(self.clean_pipelines()),
                asyncio.create_task(self.cancel_pipelines()),
                asyncio.create_task(self.query_prometheus()),
                asyncio.create_task(self.logging_protocol())
            ]
        )

    @safe_loop()
    async def scheduler_heartbeat(self):
        """Scheduler Heartbeat.

        Logs number of tasks and pipelines priodically.
        """
        await asyncio.sleep(config.heartbeat)
        self._tasks = [task for task in self._tasks if not task.done()]
        self.logger.debug(
            "Tasks %s, Pipelines %s", len(self._tasks), len(self.pipelines)
        )

    @safe_loop()
    async def logging_protocol(self):
        """Receive log lines in the ZMQ log address and
        logs them to the console.
        """
        topic, msg = await self.log_router.recv_multipart()
        msg = json.loads(msg.decode("utf8"))
        record = logging.makeLogRecord(msg)
        self.logger.handle(record)

    @safe_loop()
    async def query_prometheus(self):
        await asyncio.sleep(config.heartbeat)

        url = config.prometheus
        keys = [
            "kube_pod_container_resource_requests_cpu_cores",
            "kube_pod_container_resource_requests_memory_bytes",
            "machine_cpu_cores",
            "machine_memory_bytes",
        ]
        if not url:
            return True

        self.logger.debug("Querying Prometheus")
        url = f"{url}/api/v1/query"
        for query in keys:
            params = {"query": query}
            try:
                async with timeout(config.heartbeat):
                    async with self.session.get(url, params=params) as resp:
                        val = await resp.json()
            except ClientConnectorError:
                self.logger.error("Unable to connect to Prometheus at %s", url)
                return
            except asyncio.TimeoutError:
                self.logger.error("Prometheus request took too long. Cancelled")
                return

            val = val["data"]["result"]
            m = sum(float(item["value"][1]) for item in val)
            self.kube_metrics[query] = m
            self.logger.debug("Metric %s : %s", query, m)

    @safe_loop()
    async def load_pipelines(self):
        """Pipeline loader. Query pipelines from API and starts new."""
        await asyncio.sleep(config.heartbeat)

        # We can set maintenance mode at runtime
        if Path("/var/run/owl/nopipe").exists():
            self.logger.debug("Maintenaince mode. Pipelines not started.")
            return

        # Maximum number of pipelines at runtime
        if (maxpipe := Path("/var/run/owl/maxpipe")).exists():
            with maxpipe.open() as fh:
                self._max_pipe = int(fh.read())

        # Check for pending pipelines
        root = "/api/pipeline/list/pending"
        url = f"http://{self.env.OWL_API_SERVICE_HOST}:{self.env.OWL_API_SERVICE_PORT}{root}"
        headers = {"Authentication": f"owl {self._token}"}

        self.logger.debug("Checking for pipelines")
        try:
            async with timeout(config.heartbeat):
                async with self.session.get(url, headers=headers) as resp:
                    pipelines = await resp.json()
        except ClientConnectorError:
            self.logger.error("Unable to connect to API at %s", url)
            return
        except asyncio.TimeoutError:
            self.logger.error("API request took too long. Cancelled")
            return

        if "detail" in pipelines:
            self.logger.error(pipelines["detail"])
            return

        for pipe in pipelines:
            if pipe["id"] in self.pipelines:
                self.logger.debug("Pipeline already in list %s", pipe["id"])
                continue
            if len(self.pipelines) >= self._max_pipe:
                self.logger.info(
                    "Maximum number of pipelines reached (%s)", self._max_pipe
                )
                break

            await self.start_pipeline(pipe)

    @safe_loop()
    async def cancel_pipelines(self):
        """Query for pipelines to cancel."""
        await asyncio.sleep(config.heartbeat)

        # Check for pending pipelines
        root = "/api/pipeline/list/to_cancel"
        url = f"http://{self.env.OWL_API_SERVICE_HOST}:{self.env.OWL_API_SERVICE_PORT}{root}"
        headers = {"Authentication": f"owl {self._token}"}

        self.logger.debug("Checking for pipelines to cancel")
        try:
            async with timeout(config.heartbeat):
                async with self.session.get(url, headers=headers) as resp:
                    pipelines = await resp.json()
        except ClientConnectorError:
            self.logger.error("Unable to connect to API at %s", url)
            return
        except asyncio.TimeoutError:
            self.logger.error("API request took too long. Cancelled")
            return

        if "detail" in pipelines:
            self.logger.error(pipelines["detail"])
            return

        for pipe in list(pipelines):
            if pipe["id"] in self.pipelines:
                await self.stop_pipeline(pipe["id"], "CANCELLED")
            else:
                await self.update_pipeline(pipe["id"], "CANCELLED")

    async def stop(self):
        """Stop scheduler.

        One by one, cancel all tasks and pipelines running, in this order:

        - Pipelines. Their status is set back to PENDING.
        - Scheduler tasks.
        """
        # TODO: what we want really is that jobs are not stopped when the scheduler
        # is restarted. Save the status to a file and load it again.
        self.logger.debug("Stopping scheduler")
        if self.is_started:
            for uid in self.pipelines:
                await self.stop_pipeline(uid, "PENDING")
            # self.logger.info("Saving pipelines status")
            # with open("/var/run/owl/pipelines.pkl", "w") as fh:
            #     pickle.dump(self.pipelines, fh)

            for task in self._tasks:
                task.cancel()
                with suppress(asyncio.CancelledError):
                    await task

            await self.session.close()
            self.is_started = False

        with suppress(Exception):
            self.pipe_router.close(linger=0)
            self.log_router.close(linger=0)
            self.live_router.close(linger=0)

        self.logger.info("Owl scheduler succesfully stopped")

    async def _setup_sockets(self):
        """Setup ZMQ sockets.

        We create the sockets for bidirectional communication with
        the pipelines (ROUTER-DEALER) and for receiving the logs
        (PUB-SUB).

        These are TCP sockets listening in the ports configured by
        environmental variables `OWL_SERVICE_PORT_PIPE` and
        `OWL_SERVICE_PORT_LOGS`.
        """
        self.ctx = zmq.asyncio.Context()

        # pipelines -- used for bidirectional messages with the pipelines
        self.pipe_addr = f"tcp://0.0.0.0:{self.env.OWL_SCHEDULER_SERVICE_PORT_PIPE}"
        self.pipe_router = self.ctx.socket(zmq.ROUTER)
        self.pipe_router.set(zmq.ROUTER_HANDOVER, 1)
        self.logger.debug("Pipeline router address: %s", self.pipe_addr)
        self.pipe_router.bind(self.pipe_addr)

        # logging -- to display logs from the API and Pipelines
        self.log_addr = f"tcp://0.0.0.0:{self.env.OWL_SCHEDULER_SERVICE_PORT_LOGS}"
        self.log_router = self.ctx.socket(zmq.SUB)
        # self.log_router.setsockopt(zmq.SUBSCRIBE, b"API")
        self.log_router.setsockopt(zmq.SUBSCRIBE, b"PIPELINE")
        self.log_router.bind(self.log_addr)
        self.logger.debug("Logs router address: %s", self.log_addr)
        await asyncio.sleep(0)

    async def _start_liveness_probe(self):
        # liveness and readiness probes
        self.live_addr = f"tcp://0.0.0.0:{8080}"
        self.live_router = self.ctx.socket(zmq.REP)
        self.live_router.bind(self.live_addr)
        self.logger.debug("Liveness probe address: %s", self.live_addr)

    async def start_pipeline(self, pipe: Dict[str, Any]):
        """Start a pipeline.

        Parameters
        ----------
        pipe
            pipeline definition
        """
        uid = pipe["id"]
        user = pipe["user"]
        pipe_config = pipe["config"]

        pdef = await self.get_pipedef(pipe_config["name"])
        if pdef is None:
            self.logger.error(
                "Failed to obtain fingerprint for pipeline %s", config["name"]
            )
            raise Exception("Fingerprint failed")

        # TODO: check that resources requested can be allocated

        self.logger.debug("Starting pipeline ID %s", uid)

        with open(f"/var/run/owl/conf/pipeline_{uid}.yaml", "w") as fh:
            fh.write(json.dumps(pipe_config))

        await self._tear_pipeline(uid)

        command = "owl-server pipeline"

        env_vars = {
            "UID": uid,
            "JOBID": uid,
            "USER": user,
            "LOGLEVEL": config.loglevel,
            "PIPEDEF": json.dumps(pipe_config),
            "DASK_IMAGE_SPEC": self.env.OWL_IMAGE_SPEC,
            "OWL_IMAGE_SPEC": self.env.OWL_IMAGE_SPEC,
            "EXTRA_PIP_PACKAGES": pdef["extra_pip_packages"],
            "OMP_NUM_THREADS": "1",
            "OPENBLAS_NUM_THREADS": "1",
            "MKL_NUM_THREADS": "1",
            "VECLIB_MAXIMUM_THREADS": "1",
            "NUMEXPR_NUM_THREADS": "1",
            "NUMEXPR_MAX_THREADS": "1",
        }

        jobname = f"pipeline-{uid}"
        self.logger.debug("Creating job %s", jobname)
        status = await k8s.kube_create_job(
            jobname,
            self.env.OWL_IMAGE_SPEC,
            command=command,
            namespace=self.namespace,
            extraConfig=config.pipeline,
            env_vars=env_vars,
            service_account_name=config.pipeline["serviceAccountName"],
        )

        heartbeat = {"status": "STARTING"}
        self.pipelines[uid] = {
            "last": time.monotonic(),
            "heartbeat": heartbeat,
            "job": status,
            "name": jobname,
        }
        await self.update_pipeline(uid, heartbeat["status"])

        self._tasks.append(asyncio.create_task(self.heartbeat_pipeline(uid)))

        # ************
        logfile = f"/var/run/owl/logs/pipeline_{uid}.log"
        handler = PipelineFileHandler(logfile)
        formatter = self.logger.handlers[1].formatter
        logfilter = LogFilter(topic="PIPELINE", jobid=f"{uid}")
        handler.setFormatter(formatter)
        handler.addFilter(logfilter)
        self.logger.addHandler(handler)

        return status

    async def stop_pipeline(self, uid: int, status: str):
        """Stop pipeline

        Parameters
        ----------
        uid : int
            Unique ID of pipeline
        status : str
            status of the pipeline
        """
        self.logger.debug(f"Stopping pipeline ID {uid}")
        await self.update_pipeline(uid, status)
        if uid in self.pipelines:
            await self._tear_pipeline(uid)

    async def _tear_pipeline(self, uid: int):
        if uid not in self.pipelines:
            await asyncio.sleep(0)
            return
        with suppress(Exception):
            jobname = self.pipelines[uid]["name"]
            await k8s.kube_delete_job(jobname, self.namespace)
        self.pipelines.pop(uid)

    @safe_loop()
    async def heartbeat_pipeline(self, uid):
        """Check pipeline status.

        Send a periodic heartbeat status request to all running pipelines
        and act on result. Pipelines that report finished status are removed
        and those not responding the heartbeat request cancelled.
        """
        if uid not in self.pipelines:
            return True
        if not self.pipelines[uid].get("received", True):
            self.logger.debug("Waiting for heartbeat from pipeline %s", uid)
            return
        self.logger.debug("Sending heartbeat to pipeline %s", uid)
        await self.pipe_router.send_multipart([str(uid).encode("utf-8"), b"heartbeat"])
        self.pipelines[uid]["received"] = "False"
        await asyncio.sleep(config.heartbeat)

    @safe_loop()
    async def status_pipelines(self):
        uid, msg = await self.pipe_router.recv_multipart()
        uid, msg = int(uid.decode()), json.loads(msg.decode())
        self.logger.debug("Received heartbeat from pipeline %s : %s", uid, msg)
        self.pipelines[uid].update(
            {"last": time.monotonic(), "heartbeat": msg, "received": True}
        )
        await self.update_pipeline(uid, msg["status"])

    @safe_loop()
    async def clean_pipelines(self):
        for pipe in list(self.pipelines):  # list -> avoid dictionary change
            last = self.pipelines[pipe]["last"]
            status = self.pipelines[pipe]["heartbeat"]["status"]
            if status in ["FINISHED"]:
                self.logger.debug("Pipeline completed %s", pipe)
                await self.stop_pipeline(pipe, status)
            elif status in ["ERROR"]:
                self.logger.debug("Stopping pipeline %s with status %s", pipe, status)
                await self.stop_pipeline(pipe, status)
            elif time.monotonic() - last > 5 * config.heartbeat:
                self.logger.debug("Heartbeat not received. Stopping pipeline %s", pipe)
                status = "ERROR"
                await self.stop_pipeline(pipe, status)
        await asyncio.sleep(config.heartbeat)

    async def update_pipeline(self, uid: int, status: str):
        """Update pipeline status.

        Parameters
        ----------
        uid
            pipeline id
        status
            status to set the pipeline to
        response
            heartbeat response
        """
        await asyncio.sleep(0)

        self.logger.info("Updating pipeline %s - %s", uid, status)
        data = {"status": status}
        root = f"/api/pipeline/update/{uid}"
        url = f"http://{self.env.OWL_API_SERVICE_HOST}:{self.env.OWL_API_SERVICE_PORT}{root}"
        headers = {"Authentication": f"owl {self._token}"}

        try:
            async with timeout(config.heartbeat):
                async with self.session.post(url, json=data, headers=headers) as resp:
                    msg = await resp.json()
        except ClientConnectorError:
            self.logger.error("Unable to connect to API at %s", url)
            return
        except asyncio.TimeoutError:
            self.logger.error("API request took too long. Cancelled")
            return

        if "detail" in msg:
            self.logger.error(msg["detail"])
            return

    async def get_pipedef(self, name: str):
        """Retrieve pipeline definion file from API

        Parameters
        ----------
        name
            name of the pipeline
        """
        root = f"/api/pdef/get/{name}"
        url = f"http://{self.env.OWL_API_SERVICE_HOST}:{self.env.OWL_API_SERVICE_PORT}{root}"
        headers = {"Authentication": f"owl {self._token}"}

        self.logger.debug("Getting pipeline definition for %s", name)
        try:
            async with timeout(config.heartbeat // 2):
                async with self.session.get(url, headers=headers) as resp:
                    res = await resp.json()
        except ClientConnectorError:
            self.logger.error("Unable to connect to API at %s", url)
            return
        except asyncio.TimeoutError:
            self.logger.error("API request took too long. Cancelled")
            return

        if "detail" in res:
            self.logger.error("Pipeline not found %s", name)
            return

        return res



from setuptools import find_packages, setup

with open("README.md") as readme_file:
    readme = readme_file.read()

with open("requirements.txt") as requirements_file:
    requirements = [item.strip() for item in requirements_file.readlines()]

with open("classifiers.txt") as classifiers_file:
    classifiers = [item.strip() for item in classifiers_file.readlines()]

setup_requirements = ["pytest-runner", "flake8"]

test_requirements = ["coverage", "pytest", "pytest-cov", "pytest-mock"]

setup(
    author="Eduardo Gonzalez Solares",
    author_email="e.gonzalezsolares@gmail.com",
    classifiers=classifiers,
    description="Owl Pipeline Server Framework",
    entry_points={
        "console_scripts": [
            "owl-server=owl_server.cli:main",
            "owl-migrate=owl_server.migrations.migrate:main",
        ]
    },
    install_requires=requirements,
    license="GNU General Public License v3",
    long_description=readme,
    keywords="owl, pipeline, dask, kubernetes, python",
    name="owl-pipeline-server",
    packages=find_packages(include=["owl_server*"]),
    setup_requires=setup_requirements,
    test_suite="tests",
    tests_require=test_requirements,
    url="https://github.com/eddienko/owl-pipeline-server",
    version="0.6.0",
    zip_safe=False,
    python_requires=">=3.7",
)



import ipaddress
import re
import socket

import voluptuous as vo


class AttrDict(dict):
    """Attribute dictionary.

    Allows to access dictionary items by attribute.

    Parameters
    ----------
    d : Dict[Any, Any]
        input dictionary

    Examples
    --------
    >>> d = {'a': 1, 'b': 2}
    >>> ad = AttrDict(d)
    >>> ad.a
    ... 1
    """

    def __getattr__(self, item):
        return self.get(item, None)

    def __getstate__(self):
        return self

    def __setstate__(self, state):
        self.update(state)


class AttrSchema(vo.Schema):
    """Attribute schema.

    Allows to access schema items by attribute.
    """

    def __call__(self, d):
        return AttrDict(super().__call__(d))


def IpAddress():
    def checker(address):
        try:
            ipaddress.ip_address(address)
        except ValueError:
            raise vo.Invalid(f"Invalid IP address: {address}")
        return address

    return checker


def Hostname():
    def checker(name):
        try:
            address = socket.gethostbyname(name)
        except ValueError:
            raise vo.Invalid(f"Invalid IP address: {address}")
        return address

    return checker


def Choice(kind, choices):
    def checker(value):
        if value not in choices:
            raise vo.Invalid(f"{value} not valid for {kind}")
        return value

    return checker


def DockerImage():
    def checker(value):
        m = re.compile(r"([a-z]+)/([a-z0-9-):([a-z0-9.]+)?").match(value)
        if m is None:
            raise vo.Invalid(f"Invalid Docker image: {value!r}")
        return value

    return checker


def _validate_scheduler(conf):
    return conf


schema_env = AttrSchema(
    {
        vo.Required("OWL_SCHEDULER_SERVICE_HOST"): vo.Any(
            vo.All(str, IpAddress()), vo.All(str, Hostname())
        ),
        vo.Required("OWL_SCHEDULER_SERVICE_PORT_PIPE"): vo.All(
            vo.Coerce(int), vo.Range(min=1000, max=9999)
        ),
        vo.Required("OWL_SCHEDULER_SERVICE_PORT_LOGS"): vo.All(
            vo.Coerce(int), vo.Range(min=1000, max=9999)
        ),
        vo.Required("OWL_API_SERVICE_HOST"): vo.Any(
            vo.All(str, IpAddress()), vo.All(str, Hostname())
        ),
        vo.Required("OWL_API_SERVICE_PORT"): vo.All(
            vo.Coerce(int), vo.Range(min=1000, max=9999)
        ),
        vo.Required("OWL_IMAGE_SPEC"): vo.All(str, vo.Length(min=6)),
    }
)

schema_server = AttrSchema(
    vo.All(
        {
            vo.Required("token"): vo.All(str, vo.Length(min=20)),
            vo.Required("loglevel"): vo.All(
                str, Choice("loglevel", ["DEBUG", "INFO", "WARNING", "ERROR"])
            ),
            vo.Required("heartbeat"): vo.All(vo.Coerce(int), vo.Range(min=10, max=100)),
            vo.Required("max_pipelines"): vo.All(
                vo.Coerce(int), vo.Range(min=1, max=100)
            ),
            vo.Required("dbi"): vo.All(str, vo.Length(min=6)),
            vo.Required("env"): schema_env,
        },
        _validate_scheduler,
    ),
    # extra=vo.REMOVE_EXTRA,
    extra=vo.ALLOW_EXTRA,
)

schema_resources = AttrSchema(
    {
        vo.Optional("workers", default=3): vo.All(int, vo.Range(min=1, max=100)),
        vo.Optional("memory", default=7): vo.All(int, vo.Range(min=2, max=100)),
        vo.Optional("threads", default=2): vo.All(int, vo.Range(min=1, max=100)),
        vo.Optional("processes", default=1): vo.All(int, vo.Range(min=1, max=100)),
        vo.Optional("image"): DockerImage(),
        vo.Optional("dynamic", default=False): bool,
        vo.Optional("workers_min", default=1): vo.All(int, vo.Range(min=1, max=100)),
    }
)

schema_pipeline = AttrSchema(
    {vo.Required("resources"): schema_resources}, extra=vo.ALLOW_EXTRA
)



import asyncio
import functools
import sqlite3
from datetime import datetime
from typing import Any, Dict, Optional

import databases
from fastapi import FastAPI, Header, HTTPException, status
from owl_server.config import config
from pydantic import BaseModel

from .. import database as db
from ..crypto import PBKDF2PasswordHasher, get_random_string

OWL_USERNAME = "owl"

database = databases.Database(config.dbi)


def authenticate(admin=False):
    def wrapper(func):
        @functools.wraps(func)
        async def wrapped(*args, **kwargs):
            auth = kwargs["authentication"]
            auth += " dummy dummy"
            username, token, *_ = auth.split()
            await check_token(username, token)
            if admin:
                await check_admin(username)
            kwargs["username"] = username
            return await func(*args, **kwargs)

        return wrapped

    return wrapper


async def check_token(username, token) -> bool:
    q = db.Token.select().where(db.Token.c.username == username)
    res = await database.fetch_one(q)
    if (not res) or (res.token != token):
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED, detail="Unauthorized",
        )


async def check_password(user) -> bool:
    q = db.User.select().where(db.User.c.username == user.username)
    res = await database.fetch_one(q)
    if not res:
        return False
    hasher = PBKDF2PasswordHasher()
    return hasher.verify(user.password, res.password)


async def check_admin(username) -> bool:
    q = db.User.select().where(db.User.c.username == username)
    res = await database.fetch_one(q)
    if (not res) or (res.is_admin is False):
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED, detail="Unauthorized",
        )


async def check_config(config):
    await asyncio.sleep(0)
    if not config:
        raise HTTPException(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY, detail="Empty config",
        )
    return config


async def check_status(st):
    await asyncio.sleep(0)
    if st not in [
        "PENDING",
        "STARTING",
        "RUNNING",
        "FINISHED",
        "ERROR",
        "TO_CANCEL",
        "CANCELLED",
    ]:
        raise HTTPException(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY, detail="Invalid status",
        )
    return st


app = FastAPI()


@app.on_event("startup")
async def startup():
    await database.connect()
    token = config.token
    try:
        q = db.Token.insert().values(username=OWL_USERNAME, token=token)
        await database.execute(q)
    except sqlite3.IntegrityError:
        q = (
            db.Token.update()
            .where(db.Token.c.username == OWL_USERNAME)
            .values(token=token)
        )
        await database.execute(q)


@app.on_event("shutdown")
async def shutdown():
    await database.disconnect()


class User(BaseModel):
    username: str
    password: Optional[str] = None
    is_admin: Optional[bool] = False


class Pipeline(BaseModel):
    config: Optional[Dict[str, Any]] = None
    status: Optional[str] = None


class PipeDef(BaseModel):
    name: str
    pdef: str
    extra_pip_packages: str


@app.get("/api/pipeline/list/{status}")
@authenticate()
async def pipeline_list(status: str, authentication=Header(None), username=None):
    st = await check_status(status.upper())
    q = db.Pipeline.select().where(db.Pipeline.c.status == st)
    res = await database.fetch_all(q)
    return res


@app.get("/api/pipeline/status/{uid}")
@authenticate()
async def pipeline_status_get(uid: int, authentication=Header(None), username=None):
    q = db.Pipeline.select().where(db.Pipeline.c.id == uid)
    res = await database.fetch_one(q)
    return res


@app.post("/api/pipeline/add")
@authenticate()
async def pipeline_add(pipe: Pipeline, authentication=Header(None), username=None):
    config = await check_config(pipe.config)
    q = db.Pipeline.insert().values(
        user=username, config=config, created_at=datetime.now(), status="PENDING"
    )
    uid = await database.execute(q)
    return {"id": uid}


@app.post("/api/pipeline/cancel/{uid}")
@authenticate()
async def pipeline_cancel(uid: int, authentication=Header(None), username=None):
    q = db.Pipeline.select().where(db.Pipeline.c.id == uid)
    res = await database.fetch_one(q)
    if res.user != username:
        q = db.User.select().where(db.User.c.username == username)
        res = await database.fetch_one(q)
        if not res.is_admin:
            raise HTTPException(
                status_code=status.HTTP_401_UNAUTHORIZED, detail="Unauthorized",
            )
    q = db.Pipeline.update().where(db.Pipeline.c.id == uid).values(status="TO_CANCEL")
    await database.execute(q)
    return {"id": uid, "status": "TO_CANCEL"}


@app.post("/api/pipeline/update/{uid}")
@authenticate()
async def pipeline_update(
    uid: int, pipe: Pipeline, authentication=Header(None), username=None
):
    new = await check_status(pipe.status.upper())
    q = db.Pipeline.select().where(db.Pipeline.c.id == uid)
    res = await database.fetch_one(q)
    if not res:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail=f"Pipeline {uid} not found",
        )
    if username not in [OWL_USERNAME, res.user]:
        await check_admin(username)
    update = False
    if (res.status in ["TO_CANCEL"]) and (new == "CANCELLED"):
        update = True
    elif res.status not in ["TO_CANCEL"]:
        update = True
    if update:
        q = db.Pipeline.update().where(db.Pipeline.c.id == uid).values(status=new)
        await database.execute(q)

    return {"id": uid, "status": new, "user": res.user}


@app.post("/api/auth/login")
async def login(user: User):
    res = await check_password(user)
    if not res:
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Incorrect username or password",
        )
    token = get_random_string(64)
    try:
        q = db.Token.insert().values(username=user.username, token=token)
        await database.execute(q)
    except sqlite3.IntegrityError:
        q = (
            db.Token.update()
            .where(db.Token.c.username == user.username)
            .values(token=token)
        )
        await database.execute(q)
    return {"token": token}


@app.post("/api/auth/user/add")
@authenticate(admin=True)
async def add_user(user: User, authentication=Header(None), username=None):
    hasher = PBKDF2PasswordHasher()
    salt = hasher.salt()
    password = hasher.encode(user.password, salt)
    try:
        q = db.User.insert().values(
            username=user.username, password=password, is_admin=user.is_admin
        )
        await database.execute(q)
    except sqlite3.IntegrityError:
        q = (
            db.User.update()
            .where(db.User.c.username == user.username)
            .values(password=password, is_admin=user.is_admin)
        )
        await database.execute(q)

    return {"user": user.username}


@app.post("/api/auth/user/update")
@authenticate(admin=True)
async def update_user(user: User, authentication=Header(None), username=None):
    data = {"is_admin": user.is_admin}
    if user.password:
        hasher = PBKDF2PasswordHasher()
        salt = hasher.salt()
        password = hasher.encode(user.password, salt)
        data.update({"password": password})

    q = db.User.update().where(db.User.c.username == user.username).values(**data)
    await database.execute(q)

    q = db.Token.delete(db.Token.c.username == user.username)
    await database.execute(q)

    return {"user": user.username}


@app.post("/api/auth/user/delete")
@authenticate(admin=True)
async def delete_user(user: User, authentication=Header(None), username=None):
    q = db.User.delete().where(db.User.c.username == user.username)
    await database.execute(q)

    q = db.Token.delete(db.Token.c.username == user.username)
    await database.execute(q)

    return {"user": user.username}


@app.post("/api/pdef/add")
@authenticate(admin=True)
async def add_pdef(pdef: PipeDef, authentication=Header(None), username=None):
    try:
        q = db.PipelineDefinition.insert().values(
            name=pdef.name, pdef=pdef.pdef, extra_pip_packages=pdef.extra_pip_packages
        )
        await database.execute(q)
        action = "created"
    except sqlite3.IntegrityError:
        q = (
            db.PipelineDefinition.update()
            .where(db.PipelineDefinition.c.name == pdef.name)
            .values(pdef=pdef.pdef, extra_pip_packages=pdef.extra_pip_packages)
        )
        await database.execute(q)
        action = "updated"
    return {"name": pdef.name, "action": action}


@app.get("/api/pdef/list")
@authenticate()
async def list_pdef(authentication=Header(None), username=None):
    q = db.PipelineDefinition.select()
    res = await database.fetch_all(q)
    return res


@app.get("/api/pdef/get/{name}")
@authenticate()
async def get_pdef(name: str, authentication=Header(None), username=None):
    q = db.PipelineDefinition.select().where(db.PipelineDefinition.c.name == name)
    res = await database.fetch_one(q)
    return res


@app.get("/")
def read_root():
    return {"Hello": "World"}



import asyncio
import logging
import logging.config
import os
import signal
from argparse import Namespace
from pathlib import Path

from owl_server.config import config
from owl_server.daemon import Pipeline
from owl_server.log import initlog
from owl_server.schema import schema_pipeline
from owl_server.utils import read_config

logger = logging.getLogger("owl.cli")


def run_pipeline(args: Namespace) -> None:
    """Run scheduler.

    Parameters
    ----------
    arg
        Argparse namespace containing command line flags.
    """
    initlog("pipeline")

    pdef = read_config(args.conf, schema_pipeline)

    try:
        start_loop(pdef)
    except:  # noqa
        logger.critical("Error starting the pipeline", exc_info=True)


def start_loop(pdef):
    loop = asyncio.get_event_loop()
    logger.info("Starting Pipeline")
    pipeline = Pipeline(pdef)

    loop.add_signal_handler(signal.SIGTERM, loop.stop)
    loop.add_signal_handler(signal.SIGINT, loop.stop)

    loop.run_forever()
    loop.run_until_complete(pipeline.stop())

    loop.close()
